WRAPPER_SCRIPT_TYPE="$WRAPPER_SCRIPT_TYPE_DOCKER_COMPOSE"

# Utility method to print if user types -h
# Failure conditions: none
# Parameters: none
# Depends on global: DEFAULT_ROOT_DATA_DIR
# Updates global: none
# Returns: NA

docker_usage() { 
    cat << END_USAGE
$0 - This script prepares directories on the local host for mounting into application containers

Usage: $0 <options>

Supported options
-h|--help       : Show this usage

Examples
* Start : sudo $0

END_USAGE
    exit 1
}

showDockerExamples() {
cat << END_USAGE
\033[1mExamples:\033[0m
cd ${COMPOSE_HOME}

$(docker_hook_productSpecificComposeHelp)
$(docker_postgresSpecificComposeHelp)
start:               docker-compose -p ${JF_PROJECT_NAME} up -d
stop:                docker-compose -p ${JF_PROJECT_NAME} down

\033[1mNOTE:\033[0m The compose file uses several environment variables from the .env file. Remember to run from within the [${COMPOSE_HOME}] folder

END_USAGE
}

docker_hook_copyComposeFile(){
    logSilly "Method ${FUNCNAME[0]}"
}

docker_setupModeSpecificVar() {
    # Set JF_ROOT_DATA_DIR
    if [ -z "${JF_ROOT_DATA_DIR}" ]; then
        JF_ROOT_DATA_DIR=${DEFAULT_ROOT_DATA_DIR}
    fi

    if [ -z "${SERVER_URL}" ] && [ ! -z "${JF_SHARED_JFROGURL}" ]; then
        SERVER_URL="${JF_SHARED_JFROGURL}"
    fi

    SYSTEM_YAML_FILE="${JF_ROOT_DATA_DIR}/var/etc/${FILE_NAME_SYSTEM_YAML}"
    JF_SYSTEM_YAML="${SYSTEM_YAML_FILE}"
    docker_setInstallerStateFile "${JF_ROOT_DATA_DIR}"

    COMPOSE_TEMPLATES="$COMPOSE_HOME/templates"
    COMPOSE_FILE="${COMPOSE_HOME}/docker-compose.yaml"

    if [ -z "${PROJECT_ROOT_FOLDER}" ]; then
        PROJECT_ROOT_FOLDER=${JF_ROOT_DATA_DIR##*/}
    fi
    logDebug "PROJECT_ROOT_FOLDER is [${PROJECT_ROOT_FOLDER}]"

}

# Process command line options. See docker_usage for supported options
docker_processOptions() {
    #Determine the data folder to use for artifactory
    DEFAULT_ROOT_DATA_DIR="${JF_ROOT_DATA_DIR}"
    JF_NODE_TYPE="$NODE_TYPE_STANDALONE"

    while [[ $# > 0 ]]; do
        case "$1" in
            -v | --verbose)
                VERBOSE_MODE=true
                logger "Running installer in verbose mode"
                shift
            ;;
            -h | --help)
                docker_usage
                exit 0
            ;;
            *)
                logError "Option is not supported!"
                docker_usage
            ;;
        esac
    done
}



# Utility method to check if an application is already running
# Failure conditions:
## Exits if compose file is not available
# Parameters: none
# Depends on global: COMPOSE_HOME, COMPOSE_FILE, PROJECT_NAME
# Updates global: none
# Returns: 0 if running. 1 if not running

docker_isDeployed () {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    pushd $COMPOSE_HOME > /dev/null
    [ -f ${COMPOSE_FILE} ] || errorExit "Cannot check if system is deployed. ${COMPOSE_FILE} is missing"
    logger "Checking if the application is running. Executing command [docker-compose -p ${JF_PROJECT_NAME} ps -q]"
    docker-compose -p ${JF_PROJECT_NAME} ps -q > /dev/null || exitCode=$? && true
    local is_deployed="${exitCode}"
    popd > /dev/null
    if [ -n "$is_deployed" ]; then
        logDebug "Application is deployed: [true]"
        return 0
    else
        logDebug "Application is deployed: [false]"
        return 1
    fi
}

docker_hook_productSpecificComposeHelp(){
    logSilly "Method ${FUNCNAME[0]}"
}

# Utility method to display a successMessage after a successful dockercompose start
# Failure conditions:NA 
# Parameters: none
# Depends on global: JF_ROOT_DATA_DIR, EXTRA_MSG, 
# Updates global: none
# Returns: NA

docker_showSuccessMessage() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    local contentToPrint=$(cat << CONTENT_TO_PRINT
Installation directory: [$JF_ROOT_DATA_DIR] contains data and configurations.

Use docker-compose commands to start the application. Once the application has started, it can be accessed at [${SERVER_URL}]

$(showDockerExamples)

CONTENT_TO_PRINT
)
    banner "Docker setup complete" "${contentToPrint}"
}

docker_runComposeUp(){
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    pushd $COMPOSE_HOME > /dev/null
    logger "Running command [docker-compose -p ${JF_PROJECT_NAME} -f ${COMPOSE_FILE} up -d --remove-orphans]"
    docker-compose -p ${JF_PROJECT_NAME} -f ${COMPOSE_FILE} up -d --remove-orphans || errorExit "Failed to start application services"
    popd > /dev/null
}

# Utility method to start a single container
# Failure conditions: if start fails
# Parameters: 
    # Name of the service (NOTE: *not* container name)
# Depends on global: JF_PROJECT_NAME, COMPOSE_FILE
# Updates global: none
# Returns: NA

docker_startContainer() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    pushd $COMPOSE_HOME > /dev/null
        docker-compose -f ${COMPOSE_FILE} -p ${JF_PROJECT_NAME} up -d "$1" ||  errorExit "Failed to start $1"
    popd > /dev/null
}

# Utility method to copy a file to a container
# Failure conditions: if copying fails
# Parameters: 
    # Name of the file
    # Name of the container
    # Destination folder
# Depends on global: none
# Updates global: none
# Returns: NA

docker_copyFileToContainer() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    local fileToCopy="$1"
    local destinationContainer="$2"
    local destinationFolder="$3"
    # Copy script to container
    [ -f "$fileToCopy" ] || errorExit "Could not copy file. [$fileToCopy] is missing"

    local n=1
    local max=3
    local delay=2
    local SUCCESS=false

    while [[ $n -lt $max && "$SUCCESS" == "false" ]] ; do
        docker cp "${fileToCopy}" ${destinationContainer}:${destinationFolder} && SUCCESS=true || SUCCESS=false
        ((n++))
        [[ $n -lt $max && "$SUCCESS" == "false" ]] && logger "Retrying copy of ${fileToCopy} to ${destinationContainer}" || true
        sleep $delay;
    done
    if [[ "$SUCCESS" == "true" ]]; then
        logger "Successfully copied [$fileToCopy] to [${destinationContainer}]"
    else
        errorExit "Failed to copy [$fileToCopy] to [${destinationContainer}]"
    fi
}

docker_addToEnvFile(){
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    if [[ -z "$1" || "$1" == "" ]]; then
        return
    fi

    local osType=$(uname)

    local key_to_add=$1
    local value_to_add=$2
    
    local env_path="$ENV"
    
    export $key_to_add="${value_to_add}"
    
    local entryExists="$FLAG_Y"
    grep "$key_to_add=" "$env_path" >/dev/null 2>&1 || entryExists="$FLAG_N"

    if [ "$entryExists" == "$FLAG_Y" ]; then
        logDebug "replacing [$key_to_add] in $env_path"
        replaceText "$key_to_add=.*" "$key_to_add=$value_to_add" "${env_path}"
    fi
}

_getYamlValueAndUpdateEnv() {
    local sourceKey="$1"
    local targetEnv="$2"
    local sourceYaml="$3"

    logDebug "Method ${FUNCNAME[0]} sourceKey:[$sourceKey], targetEnv:[$targetEnv]"

    docker_getSystemValue "$sourceKey" "$FLAG_NOT_APPLICABLE" "$sourceYaml"

    if [ ! -z "$YAML_VALUE" ]; then
        eval $targetEnv=${YAML_VALUE}
        docker_addToEnvFile "$targetEnv" "${YAML_VALUE}"
        logDebug "The key [$targetEnv] in .env has been updated from system.yaml [$sourceKey] with value [$(getPrintableValueOfKey "$sourceKey" "$YAML_VALUE")]"
    fi
}

_commonYamlProperties() {
    _getYamlValueAndUpdateEnv "shared.jfrogUrl" "JF_SHARED_JFROGURL" "$1"
    _getYamlValueAndUpdateEnv "shared.node.id" "JF_SHARED_NODE_ID" "$1"
    _getYamlValueAndUpdateEnv "shared.node.ip" "JF_SHARED_NODE_IP" "$1"
    _getYamlValueAndUpdateEnv "shared.security.joinKey" "JF_SHARED_SECURITY_JOINKEY" "$1"
    _getYamlValueAndUpdateEnv "shared.security.masterKey" "JF_SHARED_SECURITY_MASTERKEY" "$1"
    _getYamlValueAndUpdateEnv "router.entrypoints.externalPort" "JF_ROUTER_ENTRYPOINTS_EXTERNALPORT" "$1"
}

# This method has been introduced to ensure that any pre-existing configuration is copied over 
# from the system yaml to this file
docker_updateFromYaml () {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"

    SYSTEM_YAML_FILE="${JF_ROOT_DATA_DIR}/var/etc/${FILE_NAME_SYSTEM_YAML}"

    # update from the existing system yaml file if one is available
    if [ ! -z "$SYSTEM_YAML_FILE" ] && [ -f "$SYSTEM_YAML_FILE" ]; then
        _commonYamlProperties "$SYSTEM_YAML_FILE"
    fi
}

docker_doValidations() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    logger "Validating System requirements"
    validateLinuxOrMac
    validateSudo
    checkDocker
    docker_validateDiskSpace
    logDebug "All validations successful"
}

docker_validateDiskSpace() {
    if [ ! -z "${JF_ROOT_DATA_DIR}" ]; then
        if [ -d "${JF_ROOT_DATA_DIR}/var" ]; then
            validateDiskSpace "${JF_ROOT_DATA_DIR}/var"
        fi
    fi
}
# hook implemented in each individual script if necessary
docker_hook_setDirInfo () {
    logSilly "Method ${FUNCNAME[0]}"
}

docker_setDirInfo() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    logDebug "Root Directory being used is: ${JF_ROOT_DATA_DIR}"
    #Ensures that the user's chosen root directory is set as an env variable overriding the default
    export JF_ROOT_DATA_DIR="${JF_ROOT_DATA_DIR}"
    #update the env file with the changed data folder location
    THIRDPARTY_DATA_ROOT="/var/data" #This variable is used by each bootstrap (can be ignored)
    docker_hook_setDirInfo #Implemented in each script
}

docker_setRootDirPermissions() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    local OS_TYPE=$(uname)
    local message="Aborting Installation. The folder [${JF_ROOT_DATA_DIR}] should be owned by [${DOCKER_USER}:${DOCKER_USER}]. Please fix by running the command: [chown ${DOCKER_USER}:${DOCKER_USER} ${JF_ROOT_DATA_DIR}]"

    # Fix directories ownerships
    io_setOwnershipNonRecursive "${JF_ROOT_DATA_DIR}" "${DOCKER_USER}" "${DOCKER_USER}" || errorExit "$message"

    # Give wide permissions on Mac (to support the non-root Artifactory and Nginx containers)
    if [ "${OS_TYPE}" == "Darwin" ]; then
        logDebug "Setting 777 permissions on [${JF_ROOT_DATA_DIR}]"
        chmod -R 777 "${JF_ROOT_DATA_DIR}" || errorExit "Setting 777 permissions on [${JF_ROOT_DATA_DIR}] failed"
    fi
}

docker_createThirdPartyDir() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    logger "Creating third party directories (if necessary)"

    io_createPostgresDir || errorExit "Aborting installation. Unable to create third-party folder necessary for PostgreSQL"
}

docker_hook_createDirectories() {
    logSilly "Method ${FUNCNAME[0]}"
}

docker_createDirectories() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    # Check if this is the first creation of the directories
    if [ ! -d "${JF_ROOT_DATA_DIR}" ]; then
        logDebug "First setup. Setting FIRST_SETUP=true"
        FIRST_SETUP=true #This will be used to set 777 permissions on Macs
    fi

    createDir "${JF_ROOT_DATA_DIR}/var" "${DOCKER_USER}" "${DOCKER_USER}"
    createDir "${JF_ROOT_DATA_DIR}/var/etc" "${DOCKER_USER}" "${DOCKER_USER}"
    createDir "${JF_ROOT_DATA_DIR}/var/data" "${DOCKER_USER}" "${DOCKER_USER}"
    createDir "${JF_ROOT_DATA_DIR}/var/log" "${DOCKER_USER}" "${DOCKER_USER}"
    createDir "${JF_ROOT_DATA_DIR}/var/etc/security" "${DOCKER_USER}" "${DOCKER_USER}"

    docker_hook_createDirectories
    
    logDebug "value of [FIRST_SETUP] is [$FIRST_SETUP]"
    if [ "$FIRST_SETUP" == true ]; then
        io_setOwnership "${JF_ROOT_DATA_DIR}" "${DOCKER_USER}" "${DOCKER_USER}" || errorExit "Setting ownership of [${JF_ROOT_DATA_DIR}] to [${DOCKER_USER}:${DOCKER_USER}] failed"
    fi
}

docker_readKeyFile () {
    FILE_CONTENT=""
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    
    local fileName=$1
    local sleepTimer=5
    local counter=1
    local maxTries=5

    local defaultFile="${JF_ROOT_DATA_DIR}/var/etc/security/${fileName}"
    echo ""
    echo -n "Waiting for [$defaultFile] to be created"
    
    #wait for the file to get created
    while [ ${counter} -lt ${maxTries} ]; do
        let "sleepTimer=$sleepTimer*2"
        if [ ${VERBOSE_MODE} != "true" ];then
            echo -n "."
        else
            logDebug "sleepTimer: [$sleepTimer], counter: [$counter]"
        fi
        sleep ${sleepTimer}
        if [ -f "${defaultFile}" ];then
            let "counter=$maxTries"
        else
            let "counter+=1"
        fi
        
    done
    echo ""
    if [ -f "${defaultFile}" ];then
        logDebug "${defaultFile} found"
        FILE_CONTENT=$(cat "${defaultFile}" || errorExit "Could not read ${defaultFile}" )
    else
        logDebug "${defaultFile} NOT found"
        FILE_CONTENT="NOT_SET"
    fi
    
}

docker_setSystemValue() {
    io_setYQPath

    local keyToUpdate="$1"
    local valueToUpdate="$2"

    local targetFile=${3:-$SYSTEM_YAML_FILE}
    local overwrite=${4:-"$FLAG_N"}
    logDebug "Method ${FUNCNAME[0]}. Setting [$keyToUpdate] to [$targetFile]"

    # If the key being updated already exists in the system.yaml, DONT update it: system.yaml is considered the source of truth
    if [ "${targetFile}" == "$SYSTEM_YAML_FILE" ]; then
        overwrite="$FLAG_N"
    fi

    getYamlValue "$keyToUpdate" "${targetFile}" "$FLAG_N"
    # If a value already exists, check if we need to overwrite
    if [ ! -z "${YAML_VALUE}" ] && [ "${overwrite}" == "$FLAG_N" ]; then
        return 0
    fi
    setSystemValue "$keyToUpdate" "$valueToUpdate" "$targetFile"
}

docker_removeSystemValue() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    io_setYQPath
    removeYamlValue "$1" "$2"
}

docker_getSystemValue() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    io_setYQPath
    local keyToFetch="$1"
    local fileToUpdateFrom="$3"
    local default=
    local envVar="$2"
    getYamlValue "$keyToFetch" "${fileToUpdateFrom}" "$FLAG_N"
    logDebug "Method ${FUNCNAME[0]}. key:[$keyToFetch]"
    if [ -z "${YAML_VALUE}" ] && [ ! -z "$envVar" ] && [ "$envVar" != "$FLAG_NOT_APPLICABLE" ]; then
        eval YAML_VALUE=\${${envVar}}
    fi
}

#this is a hook. can be implemented in each individual script
docker_hook_setupThirdParty () {
    logSilly "Method ${FUNCNAME[0]}"
}


# A hook that is meant to be implemented by each product to create any configuration specific to it
docker_hook_postSystemYamlCreation () {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
}

docker_installToSystemYaml() {
  logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"  
  local keyToFetch=$1
  local envVariable="$2"
  local systemYamlKey="$3"
  local defaultValue="$4"
  local targetYaml="${5:-${SYSTEM_YAML_FILE}}"

  if [ ! -f "$INSTALLER_YAML" ]; then
    logDebug "No installer yaml found. Skipping step"
    return
  fi
  docker_getSystemValue "$keyToFetch" "$envVariable" "$INSTALLER_YAML"
  local valueToUpdate=${YAML_VALUE:-"$defaultValue"}
  if [ -z "$valueToUpdate" ]; then
    return 1
  fi
  docker_setSystemValue "$systemYamlKey" "$valueToUpdate" "$targetYaml"
  return 0
}

docker_createSystemYaml() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    
    JF_SHARED_NODE_IP=${JF_SHARED_NODE_IP:-"$(io_getPublicHostIP)"}
    docker_setSystemValue "shared.node.ip" "$JF_SHARED_NODE_IP" "${SYSTEM_YAML_FILE}"
    
    docker_setSystemValue "shared.node.name" "$(io_getPublicHostName)" "${SYSTEM_YAML_FILE}"

    if [ ! -z "${JF_SHARED_JFROGURL}" ]; then
        docker_setSystemValue "shared.jfrogUrl" "${JF_SHARED_JFROGURL}" "${SYSTEM_YAML_FILE}"
    fi

    if [ ! -z "${JF_SHARED_SECURITY_JOINKEY}" ]; then
        docker_setSystemValue "shared.security.joinKey" "${JF_SHARED_SECURITY_JOINKEY}" "${SYSTEM_YAML_FILE}"
    fi

    docker_setSystemValue "router.entrypoints.internalPort" "7046" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "router.entrypoints.externalPort" "7082" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "router.entrypoints.traefikApiPort" "7049" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "router.entrypoints.grpcPort" "7047" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "catalog.mode" "singleTenant" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "catalog.central.url" "https://jfscatalogcentral.jfrog.io" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "catalog.logging.application.level" "info" "$SYSTEM_YAML_FILE"
    docker_setSystemValue "shared.node.id" "catalog-$(io_getPublicHostID)" "${SYSTEM_YAML_FILE}"

    if [ ! -z "${JF_SHARED_SECURITY_MASTERKEY}" ]; then
        docker_setSystemValue "shared.security.masterKey" "${JF_SHARED_SECURITY_MASTERKEY}" "${SYSTEM_YAML_FILE}"
    fi

    # if [ ! -z "${JF_ROUTER_ENTRYPOINTS_EXTERNALPORT}" ]; then
    #   docker_setSystemValue "router.entrypoints.externalPort" "${JF_ROUTER_ENTRYPOINTS_EXTERNALPORT}" "${SYSTEM_YAML_FILE}"
    # fi
    docker_installToSystemYaml "$KEY_DATABASE_TYPE"     "$FLAG_NOT_APPLICABLE" "${SYS_KEY_SHARED_DATABASE_TYPE}"     || logSilly "Key: KEY_DATABASE_TYPE not applicable"
    docker_installToSystemYaml "$KEY_DATABASE_DRIVER"   "$FLAG_NOT_APPLICABLE" "${SYS_KEY_SHARED_DATABASE_DRIVER}"   || logSilly "Key: KEY_DATABASE_DRIVER not applicable"    
    docker_installToSystemYaml "$KEY_DATABASE_URL"      "$FLAG_NOT_APPLICABLE" "${SYS_KEY_SHARED_DATABASE_URL}"      || logSilly "Key: KEY_DATABASE_URL not applicable"
    docker_installToSystemYaml "$KEY_DATABASE_USERNAME" "$FLAG_NOT_APPLICABLE" "${SYS_KEY_SHARED_DATABASE_USERNAME}" || logSilly "Key: KEY_DATABASE_USERNAME not applicable"
    docker_installToSystemYaml "$KEY_DATABASE_PASSWORD" "$FLAG_NOT_APPLICABLE" "${SYS_KEY_SHARED_DATABASE_PASSWORD}" || logSilly "Key: KEY_DATABASE_PASSWORD not applicable"
    # docker_installToSystemYaml "$KEY_ROUTER_ENTRYPOINTS_EXTERNALPORT" "$FLAG_NOT_APPLICABLE" "$SYS_KEY_ROUTER_ENTRYPOINTS_EXTERNALPORT" || logSilly "Key: KEY_ROUTER_ENTRYPOINTS_EXTERNALPORT not applicable"

    docker_getSystemValue "${KEY_CLUSTER_MASTER_KEY}" "${FLAG_NOT_APPLICABLE}" "${INSTALLER_YAML}"
    local masterKeyValue="${YAML_VALUE}"
    local masterKeyFile="$JF_ROOT_DATA_DIR/var/etc/security/master.key"
    if [ ! -f "${masterKeyFile}" ] && [ ! -z "${masterKeyValue}" ]; then
        createRecursiveDir "$JF_ROOT_DATA_DIR" "var/etc/security" "${DOCKER_USER}" "${DOCKER_USER}"
        echo -n "${masterKeyValue}" > "${masterKeyFile}" || warn "Could not create ${masterKeyFile}"
        io_setOwnership "${masterKeyFile}" "${DOCKER_USER}" "${DOCKER_USER}" || warn "Setting ownership of [${masterKeyFile}] to [${DOCKER_USER}:${DOCKER_USER}] failed" 
    fi
    docker_addToEnvFile "ROOT_DATA_DIR" "$ROOT_DATA_DIR"
    docker_hook_postSystemYamlCreation
    io_setOwnership "${SYSTEM_YAML_FILE}" "$DOCKER_USER" "$DOCKER_USER"
}

docker_createInstallerStateYaml(){
    local stateIndicatorValue=""

    [ ! -z "${INSTALLER_STATE_YAML_FILE}" ] || return 0
    setInstallerStateKeys
    for stateIndicatorKey in ${INSTALLER_STATE_KEYS}; do
        eval stateIndicatorValue=\$${stateIndicatorKey}
        [ ! -z "${stateIndicatorValue}" ] || continue
        docker_installToSystemYaml "$stateIndicatorValue" "$FLAG_NOT_APPLICABLE" "$stateIndicatorValue" "NOT_SET" "${INSTALLER_STATE_YAML_FILE}" \
            || logSilly "Key: $stateIndicatorKey not applicable"
    done

    io_setOwnership "${INSTALLER_STATE_YAML_FILE}" "$DOCKER_USER" "$DOCKER_USER"
}

docker_postgresSpecificComposeHelp() {
    getYamlValue "$KEY_POSTGRES_INSTALL" "$INSTALLER_YAML" "false"
    local isPostgresExternalised=$([ "$YAML_VALUE" == "$FLAG_Y" ] && echo -n "$FLAG_N" || echo -n "$FLAG_Y")
    if [[ "$isPostgresExternalised" == "$FLAG_Y" ]]; then
        return
    fi
cat << END_USAGE
start postgresql:    docker-compose -p ${JF_PROJECT_NAME}-postgres -f ${JFROG_POSTGRES_FILENAME} up -d
stop  postgresql:    docker-compose -p ${JF_PROJECT_NAME}-postgres -f ${JFROG_POSTGRES_FILENAME} down
END_USAGE
}

_backupFile() {
    local sourceFile=$1
    local destinationFile=$2
    
    if [ -f "${sourceFile}" ]; then

        yes | \cp  -f "${sourceFile}" "${destinationFile}" || warn "Could not copy [${sourceFile}] to [${destinationFile}]"
        
        io_setOwnership "${destinationFile}" "${DOCKER_USER}" "${DOCKER_USER}" || warn "Setting ownership of [${destinationFile}] to [${DOCKER_USER}:${DOCKER_USER}] failed"
    fi
}

docker_cleanup () {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    wrapper_removeInstallerYaml
    _backupFile "${SCRIPT_HOME}/config.log" "${JF_ROOT_DATA_DIR}/var/log/config.log"
    #yes | \cp  -f "${SCRIPT_HOME}"/templates/system*.yaml  "${JF_ROOT_DATA_DIR}/var/etc/"

    #local destinationFile="${JF_ROOT_DATA_DIR}/var/etc/system.basic-template.yaml"
    #io_setOwnership "$destinationFile" "${DOCKER_USER}" "${DOCKER_USER}" || warn "Setting ownership of [${destinationFile}] to [${DOCKER_USER}:${DOCKER_USER}] failed"
    
    #destinationFile="${JF_ROOT_DATA_DIR}/var/etc/system.full-template.yaml"
    #io_setOwnership "$destinationFile" "${DOCKER_USER}" "${DOCKER_USER}" || warn "Setting ownership of [${destinationFile}] to [${DOCKER_USER}:${DOCKER_USER}] failed"
}


docker_isDatabaseExternalised() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    local typeKey="$1"
    local typeValue="$2"
    local urlKey="$3"

    IS_EXTERNAL_DB="$FLAG_N"

    # Get the db type configured
    docker_getSystemValue "$typeKey" "$FLAG_NOT_APPLICABLE" "$SYSTEM_YAML_FILE"

    # If it does not match what we expect, abort
    if [ "$typeValue" != "$YAML_VALUE" ]; then
        logDebug "typeMismatch. Aborting docker_isDatabaseExternalised as unnecessary"
        return
    fi

    # Since DB entry is present in the yaml, check the url
    docker_getSystemValue "$urlKey" "$FLAG_NOT_APPLICABLE" "$SYSTEM_YAML_FILE"
    local connectionURL=$YAML_VALUE

    # IF no connectionURL is present, abort
    if [ -z "$connectionURL" ]; then
        logDebug "no connection URL. Aborting docker_isDatabaseExternalised as unnecessary"
        return
    fi

    IS_LOCALHOST="$FLAG_N";

    # Check if the database connection url has the container name
    logDebug "searching for [//$dockerComposeKey:] in [$connectionURL]"
    io_hasMatch "$connectionURL" "//$dockerComposeKey:" && {
        logDebug "Found container name"
        IS_LOCALHOST="$FLAG_Y";
    } || logDebug "Did not find match for container name [$dockerComposeKey]"

    if [ "$IS_LOCALHOST" == "$FLAG_Y" ]; then
        logDebug "Is not external DB"
        return
    fi

    # Check if the database connection url points to this machine itself
    io_getIsLocalhost "$connectionURL"
    if [ "$IS_LOCALHOST" == "$FLAG_Y" ]; then
        logDebug "Is not external DB"
        return
    fi
    
    IS_EXTERNAL_DB="$FLAG_Y"
}

docker_postMigration_hook() {
    logSilly "Method ${FUNCNAME[0]}"
}

docker_createLogFile() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    redirectLogsToFile #creates variable DESTINATION_LOG_FILE
    LOG_FILE="${SCRIPT_HOME}/config.log"
    exec 6>>$LOG_FILE
    exec &> >(tee -a "$LOG_FILE")
    setFilePermission "$LOG_FILE_PERMISSION" "$LOG_FILE"
}

# Utility method to copy the logrotate binary to the appropriate folder
_addLogRotateBinary() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    local productHome="$1"
    
    # Only Linux supported at the moment
    if [ $(uname) != "Linux" ]; then
        return
    fi

    # Make sure the binary exists in the installer
    local sourceLogRotateBinary="${COMPOSE_HOME}/third-party/logrotate/logrotate"
    [[ $(uname -m) == "arm64" ]] && sourceLogRotateBinary="${COMPOSE_HOME}/third-party/logrotate/logrotate_arm64" || true

    # TODO We can remove this block after all the code is restructured.
    if [ ! -f "$sourceLogRotateBinary" ]; then
        if [[ $(uname -m) == "arm64" ]]; then
            sourceLogRotateBinary="${SCRIPT_HOME}/third-party/logrotate/logrotate_arm64"
        else
            sourceLogRotateBinary="${SCRIPT_HOME}/third-party/logrotate/logrotate_amd64"
        fi
    fi
    [ -f "$sourceLogRotateBinary" ] || { logDebug "Could not find the file: [$sourceLogRotateBinary]"; }

    # Create the target directory & move the logrotate binary to it
    mkdir -p "$productHome/app/third-party/logrotate"
    cp "$sourceLogRotateBinary" "$productHome/app/third-party/logrotate/logrotate" || return 1

    # At this point the file should be owned by root which is good since the cron job is also owned by root
}

docker_configureLogRotation() {
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    _addLogRotateBinary "$JF_ROOT_DATA_DIR" || true # Intentional. We want to still try and setup logrotate
    configureLogRotation "${PROJECT_ROOT_FOLDER}" "$JF_ROOT_DATA_DIR" "$DOCKER_USER" "$DOCKER_USER" "$(id -un)" || true
}

# This hook can be used to query for any docker-specific questions
wrapper_hook_setInstallerDefaults() {
  logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
  local dataRootFolder=${PROJECT_ROOT_FOLDER:-"$JF_PROJECT_NAME"}
  wrapper_setSystemValue "$KEY_DATA_DIR_LOCATION" "$HOME/.jfrog/$dataRootFolder" "${FLAG_N}"
}

wrapper_hook_start_getUserInputs(){
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"
    if [ ! -z "$JF_ROOT_DATA_DIR" ]; then
        logger "$PROMPT_DATA_DIR_LOCATION found in .env. Skipping prompt"
        wrapper_setSystemValue "$KEY_DATA_DIR_LOCATION" "$JF_ROOT_DATA_DIR" "${FLAG_Y}"
    else
        wrapper_askUser "DATA_DIR_LOCATION"
    fi
    docker_getSystemValue "$KEY_DATA_DIR_LOCATION" "JF_ROOT_DATA_DIR" "$INSTALLER_YAML"
    JF_ROOT_DATA_DIR=${YAML_VALUE:-"$JF_ROOT_DATA_DIR"}
    export JF_ROOT_DATA_DIR="$JF_ROOT_DATA_DIR"
    logDebug "Method: wrapper_hook_start_getUserInputs. JF_ROOT_DATA_DIR is now [$JF_ROOT_DATA_DIR]"
}

docker_getUserInputs(){
    logDebug "Method ${FUNCNAME[0]} Caller:$(caller)"

    io_setYQPath

    wrapper_getUserInputs

    # At this point the installer.yaml will be available. Use it to set runtime variables

    docker_getSystemValue "$KEY_DATA_DIR_LOCATION" "JF_ROOT_DATA_DIR" "$INSTALLER_YAML"
    JF_ROOT_DATA_DIR=${YAML_VALUE:-"$JF_ROOT_DATA_DIR"}
    export JF_ROOT_DATA_DIR="$JF_ROOT_DATA_DIR"
    ROOT_DATA_DIR="$JF_ROOT_DATA_DIR"
    export ROOT_DATA_DIR="$ROOT_DATA_DIR"
    logDebug "Method: docker_getUserInputs. JF_ROOT_DATA_DIR is now [$JF_ROOT_DATA_DIR]"

    docker_getSystemValue "$KEY_ADD_TO_CLUSTER" "$FLAG_NOT_APPLICABLE" "$INSTALLER_YAML"
    if [ -z "$YAML_VALUE" ]; then
        JF_NODE_TYPE="$NODE_TYPE_STANDALONE"
    elif [ "$YAML_VALUE" == "$FLAG_N" ]; then
        JF_NODE_TYPE="$NODE_TYPE_STANDALONE"
    else
        JF_NODE_TYPE="$NODE_TYPE_CLUSTER_NODE"
    fi
    logDebug "Method: docker_getUserInputs. JF_NODE_TYPE is now [$JF_NODE_TYPE]"

    docker_getSystemValue "$KEY_JFROGURL" "JF_SHARED_JFROGURL" "$INSTALLER_YAML"
    JF_SHARED_JFROGURL=${YAML_VALUE:-"$JF_SHARED_JFROGURL"}
    logDebug "Method: docker_getUserInputs. JF_SHARED_JFROGURL is now [$JF_SHARED_JFROGURL]"

    docker_getSystemValue "$KEY_JOIN_KEY" "JF_SHARED_SECURITY_JOINKEY" "$INSTALLER_YAML"
    JF_SHARED_SECURITY_JOINKEY=${YAML_VALUE:-"$JF_SHARED_SECURITY_JOINKEY"}
    logDebug "Method: docker_getUserInputs. JF_SHARED_SECURITY_JOINKEY is now [$JF_SHARED_SECURITY_JOINKEY]"

    docker_getSystemValue "$KEY_CLUSTER_MASTER_KEY" "JF_SHARED_SECURITY_MASTERKEY" "$INSTALLER_YAML"
    JF_SHARED_SECURITY_MASTERKEY=${YAML_VALUE:-"$JF_SHARED_SECURITY_MASTERKEY"}
    logDebug "Method: docker_getUserInputs. JF_SHARED_SECURITY_MASTERKEY is now [$JF_SHARED_SECURITY_MASTERKEY]"

    docker_getSystemValue "$KEY_POSTGRES_INSTALL" "IS_POSTGRES_REQUIRED" "$INSTALLER_YAML"
    IS_POSTGRES_REQUIRED=${YAML_VALUE:-"$IS_POSTGRES_REQUIRED"}
    logDebug "Method: docker_getUserInputs. IS_POSTGRES_REQUIRED is now [$IS_POSTGRES_REQUIRED]"

    docker_getSystemValue "$KEY_HOST_IP" "JF_SHARED_NODE_IP" "$INSTALLER_YAML"
    JF_SHARED_NODE_IP=${YAML_VALUE:-"$JF_SHARED_NODE_IP"}
    logDebug "Method: docker_getUserInputs. JF_SHARED_NODE_IP is now [$JF_SHARED_NODE_IP]"
    
}

docker_setInstallerStateFile(){
    local rootDir="$1"
    [ ! -z "${rootDir}" ] && INSTALLER_STATE_YAML_FILE="${rootDir}/var/etc/${FILE_NAME_INSTALLER_STATE_YAML}" || true
}

docker_setMaxOpenfiles() {
    if [[ "${PRODUCT_NAME}" == "${CATALOG_LABEL}" ]]; then
        local noFileVal=$(ulimit -n)
        local minNoFileMax=100000
        if [[ "$noFileVal" != "unlimited" ]] && [[ $noFileVal -lt $minNoFileMax ]]; then
            if [[ -f "/etc/security/limits.conf" ]]; then
                addLine "root hard nofile 100000" "/etc/security/limits.conf"
                addLine "root soft nofile 100000" "/etc/security/limits.conf"
            fi
            [[ -f "/etc/pam.d/common-session" ]] && addLine "session required pam_limits.so" "/etc/pam.d/common-session"
        fi
    fi
}

docker_main() {
    LOG_BEHAVIOR_ADD_NEW_LINE="$FLAG_Y"
    setPostgresVersion
    # For backward compatibility. JF_ROOT_DATA_DIR is being renamed to ROOT_DATA_DIR
    JF_ROOT_DATA_DIR=${JF_ROOT_DATA_DIR:-"$ROOT_DATA_DIR"}
    export JF_ROOT_DATA_DIR="$JF_ROOT_DATA_DIR"
    INSTALLER_YAML=${INSTALLER_YAML:-"${SCRIPT_HOME}/${FILE_NAME_INSTALLER_YAML}"}

    docker_setInstallerStateFile "${JF_ROOT_DATA_DIR}"

    io_setYQPath
    
    docker_createLogFile
    _pauseExecution "docker_createLogFile"
    
    bannerStart "Beginning JFrog ${PRODUCT_NAME} setup"
    
    docker_setMaxOpenfiles

    #Perform validations
    docker_doValidations
    _pauseExecution "docker_doValidations"

    #Parse the flags passed in and handle them
    docker_processOptions $*
    _pauseExecution "docker_processOptions"

    docker_getUserInputs
    _pauseExecution "docker_getUserInputs"

    docker_updateFromYaml
    _pauseExecution "docker_updateFromYaml"

    # At this point, we know what the user wants. Set up variables
    docker_setupModeSpecificVar

    #Determine the folders to create    
    docker_setDirInfo
    _pauseExecution "docker_setDirInfo"

    #Create all the directories necessary
    docker_createDirectories
    _pauseExecution "docker_createDirectories"

    #setup permissions on the root folder
    docker_setRootDirPermissions
    _pauseExecution "docker_setRootDirPermissions"

    #create Third party directories
    docker_createThirdPartyDir
    _pauseExecution "docker_createThirdPartyDir"

    wrapper_setPostgresVersion
    _pauseExecution "wrapper_setPostgresVersion"
    
    docker_createSystemYaml
    _pauseExecution "docker_createSystemYaml"

    docker_createInstallerStateYaml
    _pauseExecution "docker_createInstallerStateYaml"

    docker_hook_copyComposeFile

    #Bring up and bootstrap any third-party databases. 
    #NOTE: It is important that this step happens after a system.yaml has been created. It reads the file.
    docker_hook_setupThirdParty
    _pauseExecution "docker_hook_setupThirdParty"

    docker_configureLogRotation
    _pauseExecution "docker_configureLogRotation"
    
    #only start the application if system.yaml is available
    docker_showSuccessMessage
    _pauseExecution "docker_showSuccessMessage"

    #backup the env file
    docker_cleanup

    bannerStart "Done"

    exit 0
}
## Docker-compose specific functions - End
