
testVarPermissions () {
    local srcProductData="${JF_PRODUCT_HOME}/var"
    local targetProductData="${JF_PRODUCT_HOME}/var"

    if [ -L "${srcProductData}" ]; then
        # Handle darwin and linux - option -f does not work on darwin
        targetProductData=$(readlink -f ${srcProductData} 2>/dev/null || readlink ${srcProductData} 2>/dev/null)
        [ ! -z "${targetProductData}" ] || errorExit "Cound not read link from ${srcProductData}, please make sure ${srcProductData} is a valid link (will be used to store configuration and data)"
    fi

    mkdir -p "${targetProductData}" || errorExit "Unable to create directory : ${targetProductData}"
    testDirectoryPermissions "${targetProductData}"
}

createUserAndGroup(){
    [[ -z "${PRODUCT_USER}" && -z "${DEFAULT_PRODUCT_USER}" ]] \
        && errorExit "PRODUCT_USER is needed to continue with installation, set this as an evironment variable to continue" || true

    : ${PRODUCT_USER:="${DEFAULT_PRODUCT_USER}"}
    : ${PRODUCT_GROUP:="${PRODUCT_USER}"}

    getent group "${PRODUCT_GROUP}" 1>/dev/null 2>&1
    if [ $? != 0  ]; then
      logger "Group ${PRODUCT_GROUP} doesn't exist. Creating ..."
      /usr/sbin/groupadd -r ${PRODUCT_GROUP} 1>/dev/null 2>&1 \
          || errorExit "Could not add group, command [ /usr/sbin/groupadd -r ${PRODUCT_GROUP} ]"
    else
      logger "Group $PRODUCT_GROUP already exists"
    fi

    getent passwd "${PRODUCT_USER}" 1>/dev/null 2>&1
    if [ $? != 0 ]; then
      logger "User $PRODUCT_USER doesn't exist. Creating ..."
      /usr/sbin/useradd -r -d "${JF_PRODUCT_HOME}" -s /bin/bash -m -g "${PRODUCT_GROUP}" "${PRODUCT_USER}" 1>/dev/null 2>&1 \
          || errorExit "Could not add group, command [ /usr/sbin/useradd -r -d ${JF_PRODUCT_HOME} -s /bin/bash -m -g ${PRODUCT_GROUP} ${PRODUCT_USER} ]"
      chown ${PRODUCT_USER}:${PRODUCT_USER} "${JF_PRODUCT_HOME}"
    else
      logger "User $PRODUCT_USER already exists"
    fi
}

# Any common actions to be performed at the beggining of a startup script can be maintained here
# Can be enhanced over time by adding more types and validations
# This should have only methods which does not need arguments to be passed
commonStartupActions(){
    local type="${1:-docker}"
    if [ "${type}" == "docker" ]; then
        validateDiskSpace
    fi
}

setUserGroup(){
    local userKey="${1}"
    local groupKey="${2}"
    local defaultUser="${3}"
    local defaultGroup="${4:-"${defaultUser}"}"
    local user=
    local group=

    [[ ! -z "${userKey}"      ]] || return 0;
    [[ ! -z "${groupKey}"     ]] || return 0;
    [[ ! -z "${defaultUser}"  ]] || return 0;
    [[ ! -z "${defaultGroup}" ]] || return 0;

    getSystemValue "${userKey}" "NOT_SET"
    user=$YAML_VALUE

    getSystemValue "${groupKey}" "NOT_SET"
    group=$YAML_VALUE

    # Consider user as group if ony user is set
    if [[ "$group" == "NOT_SET" && "$user" != "NOT_SET" ]]; then
        group="$user"
    fi

    [ "$user"  = "NOT_SET" ] && user=""  || true
    [ "$group" = "NOT_SET" ] && group="" || true

    # Default it to default user
    PRODUCT_USER=${user:-${defaultUser}}
    PRODUCT_GROUP=${group:-${defaultGroup}}
}

# Check if the script is running with the same user specified in system.yaml
# This can be used in scripts where you want to be sure that the current script is running with the intended user
checkUser(){
    local ymlKey="${1:-shared.user}"
    local exitOnError="${2:-${FLAG_Y}}"
    local currentUser="$USER"
    local expectedUser=

    [ ! -z "${ymlKey}" ] || return 0

    getSystemValue "${ymlKey}" "NOT_SET" "false" "${JF_SYSTEM_YAML}"
    expectedUser="${YAML_VALUE}"

    if [[ -z "${expectedUser}" || "${expectedUser}" == "NOT_SET" || -z "${currentUser}" ]]; then
        return 0;
    fi

    local errorMessage=$(cat << ERROR_MESSAGE
Found ${ymlKey} set as ${expectedUser} in ${JF_SYSTEM_YAML}, this script is started as ${currentUser}. 
This script can only be invoked by ${expectedUser} as set in ${JF_SYSTEM_YAML}.
ERROR_MESSAGE
)

    if [[ "${currentUser}" != *${expectedUser}* ]]; then
        [[ "${exitOnError}" == "${FLAG_Y}" ]] && errorExit "${errorMessage}" || return 1;
    fi
}

# Do the actual permission check and chown
checkAndSetOwnerOnDir () {
    local dirToCheck=$1
    local userToCheck=$2
    local groupToCheck=$3

    if [[ -z "${dirToCheck}" || -z "${userToCheck}" || -z "${groupToCheck}" ]]; then
        warn "A user, group and a directory is mandatory to check ownership"
        return
    fi

    if [ ! -d "${dirToCheck}" ]; then
        warn "${dirToCheck} is not a directory, skipping setting of ownership"
        return
    fi

    logger "Checking permissions on $dirToCheck"
    local STAT=( $(stat -Lc "%U %G" ${dirToCheck}) )
    local USER=${STAT[0]}
    local GROUP=${STAT[1]}

    if [[ "${USER}" != "$userToCheck" ]] || [[ "${GROUP}" != "$groupToCheck"  ]] ; then
        logger "$dirToCheck is owned by $USER:$GROUP. Setting to $userToCheck:$groupToCheck."
        chown -R "${userToCheck}":"${groupToCheck}" "${dirToCheck}" || errorExit "Setting ownership on $dirToCheck failed"
    else
        logger "$dirToCheck is already owned by $userToCheck:$groupToCheck."
    fi
}

setRpmDebDataOwnership(){
    local user=$1
    local group=${2:-$1}

    if [[ -z "${user}" || -z "${group}" ]]; then
      return 0
    fi

    local dataLink="${JF_PRODUCT_HOME}/var"
    local dataDir=$(readlink -f "${dataLink}" 2>/dev/null)

    if [ -L "${dataLink}" ]; then
      chown -h "${user}":"${group}" "${dataLink}" >&6 2>&1 \
          || errorExit "Failed to set ownership (${user}:${group}) on ${dataLink}"
    fi

    if [ -d "${dataDir}" ]; then
      io_setOwnership "${dataDir}" "${user}" "${group}" >&6 2>&1 \
          || errorExit "Failed to set ownership (${user}:${group}) on ${dataDir}"
    fi
}

changeRpmDebOwnership(){
    local productHome=${1:-${JF_PRODUCT_HOME}}
    local user=${2}
    local group=${3:-${user}}

    if [ -z "${user}" ]  || [ -z "${group}" ] || [ -z "${productHome}" ]; then
        warn "A JF_PRODUCT_HOME, user and group is mandatory to set ownership on JF_PRODUCT_HOME/app"
    fi

    createDir "${JF_PRODUCT_HOME}/var/log" "${user}" "${group}"

    io_setOwnership  "${JF_PRODUCT_HOME}/app" "${user}" "${group}" || errorExit "Setting ownership of [${JF_PRODUCT_HOME}/app] to [${user}:${group}] failed"

    if [ -d "${JF_PRODUCT_HOME}/var/etc" ]; then
        io_setOwnership  "${JF_PRODUCT_HOME}/var/etc" "${user}" "${group}" || errorExit "Setting ownership of [${JF_PRODUCT_HOME}/app] to [${user}:${group}] failed"
    fi

    if [ -d "${JF_PRODUCT_HOME}/var/bootstrap" ]; then
        io_setOwnership  "${JF_PRODUCT_HOME}/var/bootstrap" "${user}" "${group}" \
            || errorExit "Setting ownership of [${JF_PRODUCT_HOME}/var/bootstrap] to [${user}:${group}] failed"
    fi
}

checkOwnerAndExit(){
    local mainDir=${1}
    local usr=${2}
    local grp=${3:-${usr}}
    local checkName="yes"

    if [ -z "${mainDir}" ] || [ -z "${usr}" ] || [ -z "${grp}" ]; then
        warn "A directory (${mainDir}), user (${usr}) and group (${grp}) are manadtory parameters to modify ownership"
        return 0
    fi

    [ -d "${mainDir}" ] || [ -L "${mainDir}" ] || [ -f "${mainDir}" ] || return 0

    io_checkOwner "${mainDir}" "${usr}" "${grp}" "${checkName}"
    if [ $? == 1 ] ; then
        logger "##############################################################################" "ERROR"
        logger "Ownership mismatch. You can try executing following instruction and do a restart" "ERROR"
        logger "Command : chown -R ${usr}:${grp} ${mainDir}" "ERROR"
        logger "##############################################################################" "ERROR"
        exit 1
    fi
}

# Validate given directory and its first level of children are set with provided ownership
testOwnership(){
    local mainDir=${1:-${JF_PRODUCT_HOME}/var}
    local usr=${2}
    local grp=${3:-${usr}}

    if [ -z "${mainDir}" ] || [ -z "${usr}" ] || [ -z "${grp}" ]; then
        warn "A directory (${mainDir}), user (${usr}) and group (${grp}) are mandatory parameters to modify ownership"
        return 0
    fi

    [ -d "${mainDir}" ] || [ -L "${mainDir}" ] || [ -f "${mainDir}" ] || return 0

    checkOwnerAndExit "${mainDir}" "${usr}" "${grp}"

    if ls ${mainDir}/ >/dev/null 2>&1 ; then
        local dirContent=${mainDir}/*
        for item in ${dirContent}
        do
            # Skip this item if its niether directory nor link nor a file
            [ -d "${item}" ] || [ -f "${item}" ] || [ -L "${item}" ] || continue

            checkOwnerAndExit "${item}" "${usr}" "${grp}"
        done
    fi
}

# Set target directory to be owned by service
# This will expect JF_PRODUCT_HOME/var to be a link
setVarLinkOwnership(){
    local user=${1}
    local group=${2}
    local varLink=${3:-${JF_PRODUCT_HOME}/var}
    local defaultTargetVar=${4}
    local curTargetVar=
    local newTargetVar=
    local targetVar=
    local isCustomDataSet="no"

    if [ ! -z "${JF_PRODUCT_VAR}" ]; then
        newTargetVar=${JF_PRODUCT_VAR}
        isCustomDataSet="yes"
    else
        newTargetVar=${defaultTargetVar}
    fi

    if [[ -z "${user}"  || -z "${group}" || -z "${varLink}" || -z "${defaultTargetVar}" ]]; then
        errorExit "A user, group, varLink and defaultTargetVar is mandatory to set ownership on JF_PRODUCT_HOME/var"
    fi

    # Handle if a non link exists as $varLink
    if [[ ! -L "${varLink}" && ( -f "${varLink}" || -d "${varLink}" ) ]]; then
        local timestamp=$(echo "$(date '+%T')" | tr -d ":")
        local currentTime="$(date '+%Y%m%d').${timestamp}"
        local backup="${varLink}.backup.${currentTime}"
        logger "Found a file/directory named var under ${JF_PRODUCT_HOME}, this needs to be a link pointing to data directory (default : ${defaultTargetVar}). A backup of this will be created with name ${backup}"
        mv -f "${varLink}" "${backup}" || errorExit "Could not move ${varLink}, command : mv -f ${varLink} ${varLink}.backup.${currentTime}"
    fi

    # Readlink will give the same path as response if it does not exist
    curTargetVar=$(readlink -f ${varLink} 2>/dev/null)

    if [[ "${curTargetVar}" == *"${varLink}"* ]]; then
        logger "Creating a symlink from ${varLink} pointing to directory ${newTargetVar}"
        mkdir -p "${newTargetVar}"                                   || errorExit "Create directory failed, command : mkdir -p \"${newTargetVar}\""
        checkAndSetOwnerOnDir "${newTargetVar}" "${user}" "${group}"
        ln -s "${newTargetVar}" "${varLink}"                         || errorExit "Failed to link data directory to JF_PRODUCT_HOME/var, command : ln -s \"${newTargetVar}\" \"${varLink}\""
    else
        # If custom data is not set, use link that is read to create and change permission of target directory
        [[ "${isCustomDataSet}" == "no" ]] && targetVar="${curTargetVar}" || targetVar="${newTargetVar}"

        mkdir -p "${targetVar}"                || errorExit "Create directory failed, command : mkdir -p \"${targetVar}\""
        checkAndSetOwnerOnDir "${targetVar}" "${user}" "${group}"

        if [[ "${isCustomDataSet}" == "yes" && "${curTargetVar}" != "${newTargetVar}" ]]; then
            bannerImportant "${varLink} is linked to ${curTargetVar}, this will be changed to point to new location ${newTargetVar}
Please move all files and folders from ${curTargetVar} to ${newTargetVar}
Data directory can be controlled using environment variable JF_PRODUCT_VAR"

            [ -L "${varLink}" ] && rm -f "${varLink}" || true
            ln -s "${newTargetVar}" "${varLink}"      || errorExit "Failed to link data directory to JF_PRODUCT_HOME/var, command : ln -s \"${newTargetVar}\" \"${varLink}\" "
        fi
    fi
}

# Take a key and default as input,
#       Throw an log message if property is set in system.yaml
#       Set equivalent environment variable with the default value
setNodeProperty(){
    local key=$1
    local defaultValue=$2
    local value=
    local transformedKey=
    local envValue=
    local logDefaultMsg="false"

    [ ! -z "${key}" ] || return

    transformedKey=$(transformKeyToEnv $key)
    envValue="${!transformedKey}"

    getSystemValue "${key}" "NOT_SET" "$logDefaultMsg"
    value="${YAML_VALUE}"

    if [[ "${value}" == "NOT_SET" ]]; then
        value="$defaultValue"
        logger "Setting $transformedKey to $value"
        eval $transformedKey="${value}"
        [ "${key}" == "${SYS_KEY_SHARED_NODE_IP}" ] &&  modifyValue "$transformedKey" "HOST_IP" || true
        export $transformedKey
    elif [[ -z "${envValue}" && "${value}" != "NOT_SET" ]]; then
        logger "Found explicit ${key} in system.yaml with value : $value"
        eval $transformedKey="${value}"
        [ "${key}" == "${SYS_KEY_SHARED_NODE_IP}" ] &&  modifyValue "$transformedKey" "HOST_IP" || true
        export $transformedKey
    elif [[ ! -z "${envValue}" ]]; then
        logger "Found $transformedKey set as environment variable with value : $envValue"
        [ "${key}" == "${SYS_KEY_SHARED_NODE_IP}" ] &&  modifyValue "$transformedKey" "HOST_IP" || true
        export $transformedKey
    fi
}

setupNodeDetails () {
    setNodeProperty "${SYS_KEY_SHARED_NODE_ID}" "$(io_getPublicHostID)"
    setNodeProperty "${SYS_KEY_SHARED_NODE_IP}" "$(io_getPublicHostIP)"
    setNodeProperty "${SYS_KEY_SHARED_NODE_NAME}" "$(io_getPublicHostName)"
}

# Translate a value from given list of old keys to new keys
# Expects a argument with list of env mapping.
# Example ENV_MAP="""shared.node.id=HA_NODE_ID
#                    connector.extra=SERVER_XML_EXTRA_CONNECTOR"""
translateEnv() {
    local envMap="$1"
    local oldKey=
    local newKey=
    local envEquivalent=
    local oldValue=
    local newValue=

    for entry in $envMap
    do
        [[ $entry == *"="* ]] || continue;
        newKey=$(echo "${entry}" | awk -F'=' '{print $1}')
        oldKey=$(echo "${entry}" | awk -F'=' '{print $2}')

        [ ! -z "$newKey" ] || continue;
        [ ! -z "$oldKey" ] || continue;

        # Skip translation of this variable if new key is already set with a value 
        getSystemValue "${newKey}" "NOT_SET"
        newValue="${YAML_VALUE}"

        # Proceed with translation only if new key is not set - new key takes priority over old key
        if [[ "$newValue" != "NOT_SET" ]]; then 
            continue
        fi

        # allowed oldValues in a environment key are upper case alphabet, number and underscore 
        [[ "$oldKey" =~ ^[A-Z_0-9]+$ ]] || continue;

        oldValue="${!oldKey}"

        if [ ! -z "${oldValue}" ]; then
            envEquivalent=$(transformKeyToEnv "${newKey}")

            logger "Mapping value from deprecated key ${oldKey} to new key ${envEquivalent}"
            export ${envEquivalent}="${oldValue}"
        fi
    done
}

# Load environment variable specified for a given service from system.yaml
# Given a service name it will append env to get and load all the env variables in it,
# For instance, if metadata is passed, metadata.env will be loaded
#
# Expects a service name as parameter,
# Sample usage :
#       exportEnv shared
#       exportEnv metadata
exportEnv(){
    local service=$1
    local envList=
    local envPath=
    local delimiter=":"
    local key=
    local value=

    [ ! -z "${service}" ] || return 0

    envPath="${service}.env"
    getSystemValue "${envPath}" "NOT_SET" "false"
    envList="${YAML_VALUE}"

    [[ "${envList}" != "NOT_SET" ]] || return
    
    OLDIFS=$IFS
    # Use newline as the only delemiter, this is to avoid transformation of a value to JF format 
    #       resulting in failure of script in case of slash and other special characters
    IFS=$'\n'
    for entry in ${envList}
    do
        if [[ "${entry}" =~ ^JF_* ]]; then
            # Throw a warning if product related environment variable (JF_*) is added as part of env list
            warn "Environment variable starting with JF_ is not allowed in ${envPath}. Please use its equivalent yaml path to pass it to application."
            continue
        fi

        key=$(echo "${entry}" | awk -F"${delimiter}" '{print $1}')
        key="$(io_trim "${key}")"

        getSystemValue "${envPath}.${key}" "NOT_SET"
        value="${YAML_VALUE}"

        [[ "${value}" != "NOT_SET" ]] || continue

        export ${key}="${value}"
    done

    # set IFS to its default value
    IFS=$OLDIFS
}