#
# Yaml helper which can perform actions on a yaml file . A key will be searched in the order of files provided in JF_SYSTEM_YAML
# The highest priority goes to environment variables
#   If a key is parent.child, the equivalent environment variable is PARENT_CHILD

resolutionlogger(){
    local key=${1}
    local yamlValue=${2}
    local valueSource=${3}
    
    if [ -z "${key}" ] || [ -z "${yamlValue}" ] || [ -z "${valueSource}" ]; then
        return
    fi

    isKeySensitive "${key}" && yamlValue="$SENSITIVE_KEY_VALUE" || yamlValue=${yamlValue}
    logger "resolved ${key} (${yamlValue}) from ${valueSource}"
}

#######################################
# Set path to yq binary
# Globals:
#   YQ_PATH
#   YQ
# Arguments:
#   None
# Returns:
#   None
#######################################
initYQ(){

    if [[ -z "${YQ_PATH}" ]]; then
        io_setYQPath
    fi

    [[ ! -z "${YQ_PATH}" ]] || errorExit "YQ_PATH is not set, script requires a \$YQ_PATH/yq to work with system.yaml"

    local os_type=$(uname)
    local os_arch=$(uname -m)
    local os_yq_name=
    [ "${os_type}" = "Darwin" ] && osYqName=yq_darwin || osYqName=yq_linux

    if [[ "${os_type}" == "Darwin" ]]; then
        osYqName=yq_darwin
    elif [[ "${os_arch}" == "arm64" || "${os_arch}" == "aarch64" ]]; then
        osYqName=yq_linux_arm64
    else
        osYqName=yq_linux
    fi

    if [ -f "${YQ_PATH}"/"${osYqName}" ]; then
        YQ="${YQ_PATH}"/"${osYqName}"
    elif [ -f "${YQ_PATH}"/yq ]; then
        YQ="${YQ_PATH}"/yq
    else
        errorExit "yq is not found in provided YQ_PATH ($YQ_PATH)"
    fi

    [ -x "${YQ}" ] || chmod +x "${YQ}" >/dev/null 2>&1 || warn "Unable to change permission of yaml parser : ${YQ}"

    # check if yq is working
    if ! "${YQ}" -V >/dev/null 2>&1; then
        warn "Unable to execute yaml parser \"${YQ} -V\", the executable might be corrupted or might not be compatible on the current operating system. (yq is used by installer to work with system.yaml)"
    fi
}

#######################################
# Append dot to key path if not exists - for yq 4x
#######################################
appendDotToKeyPath(){
    local key=$1
    if [[ ${key:0:1} == "." ]] ; then 
      echo $key; 
    else 
      echo "."$key; 
    fi
}


#######################################
# Get value from a yaml file
# Globals:
#   YAML_VALUE
#   YQ
# Arguments:
#   key           - key to be searched
#   yamlFile      - target yaml file
#   logResolution - logs a message about the file which had this key set
# Returns:
#   None
#######################################
getYamlValue(){
    initYQ
    local key=$1
    local yamlFile=$2
    local logResolution=${3:-true}

    [ -z "${key}"      ] && errorExit "a key is mandatory to retrieve value from yaml"  || true
    [ -z "${yamlFile}" ] && errorExit "yaml file is mandatory to retrieve value"        || true

    # set yaml value to undefined if yaml file is not found
    if [[ ! -f "${yamlFile}" || ! -s "${yamlFile}" ]]; then
        YAML_VALUE=
        return
    fi
    key=$(appendDotToKeyPath "$key")
    if "${YQ}" e "${key}" "${yamlFile}" >/dev/null 2>&1; then
        YAML_VALUE=$("${YQ}" e "${key}" "${yamlFile}"  2>/dev/null)
    else
        YAML_VALUE=
        return
    fi

    if [[ "${YAML_VALUE}" == "null" ]]; then
        YAML_VALUE=
    fi

    if [[ ${logResolution} == true && ! -z "${YAML_VALUE}" ]]; then
        resolutionlogger "${key}" "${YAML_VALUE}" "${yamlFile}"
    fi
}


#######################################
# Remove key from a yaml file
# Globals:
#   YQ
# Arguments:
#   key           - key to be removed
#   yamlFile      - target yaml file
# Returns:
#   None
#######################################
removeYamlValue(){
    initYQ
    local key=$1
    local yamlFile=${2:-${JF_SYSTEM_YAML}}

    [ -z "${key}" ] && errorExit "a key is mandatory to remove key from yaml" || true

    if [[ ! -f "${yamlFile}" || ! -s "${yamlFile}" ]]; then
        return
    fi
    key=$(appendDotToKeyPath "$key")
    "${YQ}" eval -i "del(${key})" "${yamlFile}" 2>/dev/null
}

#######################################
# Perform a yaml merge
# Globals:
#   YQ
# Arguments:
#   initialYamlFile           - Original file
#   additionalYamlFile        - File to be appended with the original file
# Returns:
#   None
#######################################
yamlMerge(){
    initYQ
    local initialYamlFile="${1}"
    local additionalYamlFile="${2}"
    ${YQ} eval-all -i "select(fileIndex == 0) * select(filename == \"${additionalYamlFile}\")" "${initialYamlFile}" "${additionalYamlFile}" > /dev/null || errorExit "Failed to merge ${additionalYamlFile} to  ${initialYamlFile}"
}

#######################################
# Get equivalent environment variable of yaml key
#   If a key is parent.child, the equivalent environment variable is PARENT_CHILD
# Globals:
#   YAML_VALUE
# Arguments:
#   key           - key to be searched
#   logResolution - logs a message if key was set as an environment variable
# Returns:
#   None
#######################################
getENVValue(){
    local key=$1
    local logResolution=${2:-true}

    [ -z "${key}" ] && errorExit "a key is mandatory to retrieve value from environment" || true

    key=$(transformKeyToEnv "$key")

    YAML_VALUE="${!key}"

    if [[ ${logResolution} == true && ! -z "${YAML_VALUE}" ]]; then
        resolutionlogger "${key}" "${YAML_VALUE}" "environment variable"
    fi
}

#######################################
# Get value from a system yaml file
#   Set JF_SYSTEM_YAML with the list of files to be parsed through
# Globals:
#   YAML_VALUE
#   YQ
# Arguments:
#   key           - key to be searched
#   default       - a default value to set if none of the files have this key
#   logResolution - logs a message about the file which had this key set
# Returns:
#   None
#######################################
getSystemValue(){
    initYQ
    local key=$1
    local default=${2:-""}
    local logResolution=${3:-true}

    if [ ! -z "${4}" ]; then
        local systemYamlList=${4}
    else
        systemYamlList=${JF_SYSTEM_YAML:-"${JF_PRODUCT_HOME}/var/etc/system.yaml"}
    fi

    [ -z "${key}" ] && errorExit "a key is mandatory to retrieve value from yaml"  || true

    YAML_VALUE=

    # check if this key is set as an environment variable
    getENVValue "${key}" "${logResolution}"
    [ ! -z "${YAML_VALUE}" ] && return || true

    # check if the key is set in any of the yaml files
    for yamlFile in ${systemYamlList}
    do
        getYamlValue "${key}" "${yamlFile}" "${logResolution}"
        [ ! -z "${YAML_VALUE}" ] && return || true
    done

    if [ -z "${YAML_VALUE}" ]; then
        if [ ! -z "${default}" ]; then
            YAML_VALUE="${default}"
        else
            errorExit "${key} is not set in : ${systemYamlList}"
        fi
    fi
}

#######################################
# Get value from a system yaml file
#   Optionally set JF_SYSTEM_YAML with the target system.yaml to be updated
# Globals:
#   YAML_VALUE
#   YQ
# Arguments:
#   key           - key to be modified
#   value         - value to be set
#   yamlFile      - target system yaml file to be modified
# Returns:
#   None
#######################################
setSystemValue(){
    initYQ
    local path=$1
    local value=$2
    local ymlFile=${3:-${JF_SYSTEM_YAML}}
    local option=${4:-""}
    local displayValue=

    [ -z "${path}"    ] && errorExit "a path is mandatory to modify system yaml"                || true
    [ -z "${value}"   ] && errorExit "a value is mandatory to update a path in system yaml"     || true
    [ -z "${ymlFile}" ] && errorExit "a yaml file is mandatory to update a path in system yaml" || true

    isKeySensitive "${path}" && displayValue="$SENSITIVE_KEY_VALUE" || displayValue=${value}

    logDebug "setting ${path} with ${displayValue} in ${ymlFile}"
    path=$(appendDotToKeyPath "$path")
    if [[ ! -f "${ymlFile}" || ! -s "${ymlFile}" ]]; then
        if [[ $value == *"\""* ]]; then
            value=$value ${YQ} e -n ${option} "${path} = strenv(value)" > "${ymlFile}" || errorExit "failed to set create ${path} with ${displayValue} in ${yamlFile}, make sure its parent directory is already create and $USER user has permission to create this file under it,\ncommand used : ${YQ} e -n ${option} \"${path} = \"${value}\"\" > \"${ymlFile}\""
        else
            value=$value ${YQ} e -n ${option} "${path} = env(value)" > "${ymlFile}" || errorExit "failed to set create ${path} with ${displayValue} in ${yamlFile}, make sure its parent directory is already create and $USER user has permission to create this file under it,\ncommand used : ${YQ} e -n ${option} \"${path} = \"${value}\"\" > \"${ymlFile}\""
        fi
    else
        if [[ $value == *"\""* ]]; then
            value=$value ${YQ} eval -i ${option} "${path} = strenv(value)" "${ymlFile}" || errorExit "failed to set create ${path} with ${displayValue} in ${yamlFile}, \ncommand used : ${YQ} eval -i ${option} \"${path} = \"${value}\"\" \"${ymlFile}\""
        else
            value=$value ${YQ} eval -i ${option} "${path} = env(value)" "${ymlFile}" || errorExit "failed to set create ${path} with ${displayValue} in ${yamlFile}, \ncommand used : ${YQ} eval -i ${option} \"${path} = \"${value}\"\" \"${ymlFile}\""
        fi
    fi
}

setSystemValueNoOverride(){
    local key=$1
    local value=$2
    local yamlFile=$3
    local logMsg=${4:-false}

    # Return if any of the inputs are empty
    [[ -z "$key"      || "$key"      == "" ]] && return
    [[ -z "$value"    || "$value"    == "" ]] && return
    [[ -z "$yamlFile" || "$yamlFile" == "" ]] && return

    getSystemValue "${key}" "NOT_SET" "${logMsg}" "${yamlFile}"

    if [ ${YAML_VALUE} == "NOT_SET" ]; then
        setSystemValue "${key}" "${value}" "${yamlFile}"
    else
        [ "${logMsg}" == "true" ] && logger "Skipping update of property : ${key}" || true
    fi
}

#######################################
# Resolve any system keys in a file to its respective values from system.yaml
# Syntax of key should be of the following format ${parent.child},
# A default value can be set by using || as field separator
# Example : ${artifactory.port}
#           ${node.ip||127.0.0.1}
# Globals:
#   YAML_VALUE
# Arguments:
#   target  - location to target file with keys (example /opt/jfrog/artifactory/server.xml)
# Returns:
#   None
#######################################
resolveSystemKeys(){
    local target=$1
    local keyList=
    local nestedResolutionCount=${2:-1}
    local maxNestedResolutionCount=5
    local squareBracketsKey="_SQUARE_BRACKETS_"
    local squareBracketsReplaceValue='\[\]'
    local value=
    local defaultValue=
    local emptyKeyWord=${EMPTY_KEY_WORD:-_EMPTY_}
    local keyOrig=
    local displayValue=

    [ -z "${target}" ] && errorExit "a target file is mandatory to resolve keys from system.yaml"  || true
    [ -f "${target}" ] || errorExit "target file (${target}) not found"

    # regex supports the following examples,
    #   test.name
    #   test_name
    #   artifacts.[0].name
    #   test.test.test
    #   parent.child||sample_default_value
    local keyRegex="[a-z]*[A-Z]*[0-9]*[_\-]*\[*\]*\**\.*|*\/*"

    # support upto 32 levels of special characters - test.test.test-test.test.test_test...test
    for i in $(seq 1 5); do keyRegex="${keyRegex}${keyRegex}"; done

    keyList=$(cat "${target}" | grep -o "\${${keyRegex}}")

    if [ -z "${keyList}" ]; then
        warn "No keys to resolve in ${target}"
    fi

    for key in ${keyList}; do
        defaultValue=
        if [ ! -z "${key}" ]; then
            keyOrig="${key}"
            # convert ${key} => key
            key=${key#"\${"}
            key=${key%"}"}

            # Check if a defaut value is defined
            # parent.child||sample_default
            #   defaultValue=sample_default
            #   key=parent.child
            if [[ "${key}" = *"||"* ]]; then
                defaultValue=${key#*||}
                key=${key%||*}
            fi

            getSystemValue "${key}" "${defaultValue}"
            value="${YAML_VALUE}"

            if [[ "${value}" == "${emptyKeyWord}" ]]; then
                value=""
            fi

            isKeySensitive "${key}" && displayValue="$SENSITIVE_KEY_VALUE" || displayValue=${value}

            if [[ "${value}" == "${defaultValue}" ]]; then
                logger "resolved ${keyOrig} to default value : ${displayValue}"
            fi

            # Handle resolution of multiline string value; sed does not handle newline charaters in value
            value=$(echo "${value}" | tr -d "\n")

            if [[ "${keyOrig}" == *"relaxedPathChars"* ]] || [[ "${keyOrig}" == *"relaxedQueryChars"* ]]; then
                value=$(escapeSpecialChars "${value}")
            fi

            if [[ $(uname) == "Darwin" ]]; then
                sed -i '' -e "s,\\${keyOrig//,/\\,},${value//,/\\,},g" "${target}" || errorExit "resolving ${keyOrig} to ${displayValue} failed, command used : sed -i '' -e \"s,\\${keyOrig},${displayValue},g\" ${target}"
            else
                sed -i -e "s,\\${keyOrig//,/\\,},${value//,/\\,},g" "${target}" || errorExit "resolving ${keyOrig} to ${displayValue} failed, command used : sed -i -e \"s,\\${keyOrig},${displayValue},g\" ${target}"
            fi
        fi
    done

    keyList=$(cat "${target}" | grep -o "\${${keyRegex}}")

    if [ ! -z "${keyList}" ]; then
        if [[ ${nestedResolutionCount} -lt 5 ]]; then
            nestedResolutionCount=$(($nestedResolutionCount + 1))
            resolveSystemKeys "${target}" ${nestedResolutionCount}
        else
            local errMsg="The following list of variables ${keyList} could not be resolved after 5 tries in ${target}, this could happen \
if there were some invalid characters configured in ${JF_PRODUCT_HOME}/var/etc/system.yaml or \
if there was some kind of corruption in ${target} template"
             errorExit "${errMsg}"
        fi
    fi

    if cat "${target}" | grep "${squareBracketsKey}" >/dev/null 2>&1; then
        if [[ $(uname) == "Darwin" ]]; then
            sed -i '' -e "s,${squareBracketsKey},${squareBracketsReplaceValue},g" "${target}" || true
        else
            sed -i -e "s,${squareBracketsKey},${squareBracketsReplaceValue},g" "${target}" || true
        fi
    fi
}

escapeSpecialChars(){
    local value="$1"
    if [[ "${value}" =~ .*[\^\&\;\[\]|\{\}].* ]]; then
        [[ "$value" =~ .*\^.* ]] && value="$(sed -e 's/\^/\\\^/' <<<"$value")" || true
        [[ "$value" =~ .*\&.* ]] && value="$(sed -e 's/\&/\\\&/' <<<"$value")" || true
        [[ "$value" =~ .*\[.* ]] && value="$(sed -e 's/\[/\\\[/' <<<"$value")" || true
        [[ "$value" =~ .*\].* ]] && value="$(sed -e 's/\]/\\\]/' <<<"$value")" || true
        [[ "$value" =~ .*|.*  ]] && value="$(sed -e 's/|/\\|/' <<<"$value")"   || true
        [[ "$value" =~ .*\{.* ]] && value="$(sed -e 's/{/\\{/' <<<"$value")"   || true
        [[ "$value" =~ .*\{.* ]] && value="$(sed -e 's/}/\\}/' <<<"$value")"   || true
        [[ "$value" =~ .*\;.* ]] && value="$(sed -e 's/\;/\\\;/' <<<"$value")" || true
    fi
    regex="^\".*\"$"
    if [[ "$value" =~ $regex ]]; then
        value=$(sed -e 's/^"//' -e 's/"$//' <<<"$value")
    fi
    echo "${value}"
}
